include misc.e
include machine.e

atom help, helplines
atom fallouterr
atom pressf1
sequence infostrucs
sequence pstrucs
sequence states

states = {}
infostrucs = {}
pstrucs = {}

function AllocAndPoke(sequence bytes)
atom mem
    mem = allocate(length(bytes))
    poke(mem, bytes)
    return mem
end function

help = AllocAndPoke("ATP Assembly Mini-Debugger (c) Alexander Toresson\n" &
                    "-------------------------------------------------\n\n" &
                    "This debugger offers basic functions.\n" &
                    "Hot-keys in main view:\n" &
                    "   F1: Display this text.\n" &
                    "  Esc: Quits the debugger.\n" &
                    "Everything else will make the program continue to the next asm statement.\n" &
                    "Contact: alexander.toresson@swipnet.se\n\n" &
                    "Press any key to return\n")
helplines = 11

fallouterr = AllocAndPoke("Error: Code execution continued beyond code! Press any key to crash.\n")

pressf1 = AllocAndPoke("Press F1 to view help.\n")

global function debug_asm(sequence asm)
sequence ret, line
sequence lines
integer numline
sequence temp, temp2
integer en
atom mem
    lines = {}
    ret = ""
    line = ""
    numline = 0
    for a = 1 to length(asm) do
        line &= asm[a]
        if asm[a] = '\n' then
            numline += 1
            ret &= "push dword " & sprint(numline) & "\n"
            ret &= "call near @@dbg_debugger\n"
            ret &= line
            lines = append(lines, line)
            line = ""
        end if
    end for
    if length(line) > 0 then
        line &= "\n"
        numline += 1
        ret &= "push dword " & sprint(numline) & "\n"
        ret &= "call near @@dbg_debugger\n"
        ret &= line
        lines = append(lines, line)
        line = ""
    end if
    ret &= "push dword #FFFFFFFF\n"
    ret &= "push dword #FFFFFFFF\n"
    temp = ""
    for a = 1 to length(lines) by 20 do
        en = a + 19
        if en > length(lines) then
            en = length(lines)
        end if
        temp2 = int_to_bytes(en-a+1)
        for b = a to en do
            temp2 &= sprintf("%4d", b) & " " & lines[b]
        end for
        mem = AllocAndPoke(temp2)
        infostrucs &= mem
        temp &= int_to_bytes(mem)
    end for

    mem = AllocAndPoke(temp)
    pstrucs &= mem

    states &= AllocAndPoke(int_to_bytes(1))

--    ret &= "\n\n"
    ret &= "@@dbg_debugger:\n"
    ret &= "pushad\n"
--    ret &= "db #66\n"    -- prefix, to get pushfd instead of pushf
    ret &= "pushf\n"
    ret &= "@@dbg_redraw:\n"
    ret &= "mov eax, [#" & sprintf("%x", states[length(states)]) & "]\n"
    ret &= "cmp eax, 1\n"
    ret &= "jne near @@dbg_debugskip\n"
    ret &= "mov ebp, esp\n"
    ret &= "xor edx, edx\n"
    ret &= "mov eax, [ebp+40]\n"
    ret &= "cmp eax, #FFFFFFFF\n"
    ret &= "jne @@dbg_noerror\n"
--        ret &= "mov ebx, #FFFFFFFF\n"
--        ret &= "mov [ebx], ebx\n"
        ret &= "call near @@dbg_clrscreen\n"
        ret &= "mov ebx, #" & sprintf("%x", fallouterr) & "\n"
        ret &= "mov edi, #B8000\n"
        ret &= "@@dbg_printerrloop:\n"
            ret &= "mov eax, eax\n"                           -- asm.e has got parsing difficulties with the db command after a label, so this is necessary
            ret &= "db #80\ndb #3B\ndb #0A\n"              -- this is the opcode of the next line (commented out)
--            ret &= "cmp [ebx], byte 10\n"          -- asm.e doesn't assemble this right.
            ret &= "jz @@dbg_doneprinterr\n"                   -- done?
            ret &= "mov al, [ebx]\n"
            ret &= "mov [edi], al\n"                -- put char
            ret &= "add edi, 2\n"
            ret &= "inc ebx\n"
            ret &= "jmp @@dbg_printerrloop\n"
        ret &= "@@dbg_doneprinterr:\n"
        ret &= "call near @@dbg_waitkey\n"
        ret &= "mov ebx, #FFFFFFFF\n"
        ret &= "mov [ebx], ebx\n"
    ret &= "@@dbg_noerror:\n"
    ret &= "dec eax\n"
    ret &= "mov ecx, 20\n"
    ret &= "div ecx\n"
    ret &= "mov ebx, #" & sprintf("%x", mem) & "\n"
    ret &= "mov ebx, [ebx+eax*4]\n"
    ret &= "mov ecx, [ebx]\n"
    ret &= "add ebx, 4\n"
    ret &= "mov esi, edx\n"

--    ret &= "mov edx, esi\n"                     -- -+
--    ret &= "shl edx, 5\n"                       --  |
--    ret &= "mov edi, edx\n"                     --  |- calculate offset of line
--    ret &= "shl edx, 2\n"                       --  |
--    ret &= "add edi, edx\n"                     --  |
--    ret &= "add edi, #B8000\n"                  -- -+
    ret &= "call near @@dbg_clrscreen\n"
    ret &= "mov edi, #B8000\n"
    ret &= "mov eax, 0\n"
    ret &= "@@dbg_outerloop:\n"
        ret &= "push edi\n"

        ret &= "cmp esi, eax\n"                       -- decide on color
        ret &= "jnz @@dbg_notcurline\n"
            ret &= "mov dl, 31\n"                  -- white text, grey bkg
        ret &= "jmp @@dbg_endif\n"
        ret &= "@@dbg_notcurline:\n"
            ret &= "mov dl, 15\n"                   -- white text, black bkg
        ret &= "@@dbg_endif\n"

        ret &= "push eax\n"

        ret &= "call near @@dbg_printtext\n"

        ret &= "pop eax\n"

        ret &= "inc eax\n"

        ret &= "pop edi\n"
        ret &= "add edi, 160\n"

        ret &= "dec ecx\n"
        ret &= "cmp ecx, 0\n"
    ret &= "jnz @@dbg_outerloop\n"                      -- done?

    ret &= "mov edi, #B8C80\n"
    ret &= "push edi\n"

    ret &= "mov al, 69\n"                             -- E
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 65\n"                             -- A
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 88\n"                             -- X
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp+32]\n"
    ret &= "call near @@dbg_printhex\n"

    ret &= "add edi, 4"
    ret &= "mov al, 69\n"                             -- E
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 66\n"                             -- B
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 88\n"                             -- X
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp+20]\n"
    ret &= "call near @@dbg_printhex\n"

    ret &= "add edi, 4"
    ret &= "mov al, 69\n"                             -- E
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 67\n"                             -- C
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 88\n"                             -- X
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp+28]\n"
    ret &= "call near @@dbg_printhex\n"

    ret &= "add edi, 4"
    ret &= "mov al, 69\n"                             -- E
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 68\n"                             -- D
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 88\n"                             -- X
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp+24]\n"
    ret &= "call near @@dbg_printhex\n"

    ret &= "add edi, 4"
    ret &= "mov al, 69\n"                             -- E
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 83\n"                             -- S
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 73\n"                             -- I
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp+8]\n"
    ret &= "call near @@dbg_printhex\n"

    ret &= "pop edi\n"
    ret &= "add edi, 160\n"
    ret &= "push edi\n"

    ret &= "mov al, 69\n"                             -- E
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 68\n"                             -- D
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 73\n"                             -- I
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp+4]\n"
    ret &= "call near @@dbg_printhex\n"

    ret &= "add edi, 4"
    ret &= "mov al, 69\n"                             -- E
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 66\n"                             -- B
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 80\n"                             -- P
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp+12]\n"
    ret &= "call near @@dbg_printhex\n"

    ret &= "add edi, 4"
    ret &= "mov al, 69\n"                             -- E
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 83\n"                             -- S
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 80\n"                             -- P
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp+16]\n"
    ret &= "call near @@dbg_printhex\n"

    ret &= "add edi, 4"
    ret &= "mov al, 69\n"                             -- E
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 73\n"                             -- I
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 80\n"                             -- P
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp+36]\n"
    ret &= "call near @@dbg_printhex\n"

    ret &= "add edi, 4"
    ret &= "mov al, 69\n"                             -- E
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 70\n"                             -- F
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 76\n"                             -- L
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp]\n"
    ret &= "call near @@dbg_printhex\n"

    ret &= "pop edi\n"
    ret &= "add edi, 160\n"
    ret &= "push edi\n"

    ret &= "mov al, 67\n"                             -- C
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 70\n"                             -- F
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp]\n"
    ret &= "and eax, 1\n"
    ret &= "add eax, 48\n"
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "add edi, 4\n"
    ret &= "mov al, 80\n"                             -- P
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 70\n"                             -- F
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp]\n"
    ret &= "shr eax, 2\n"
    ret &= "and eax, 1\n"
    ret &= "add eax, 48\n"
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "add edi, 4\n"
    ret &= "mov al, 65\n"                             -- A
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 70\n"                             -- F
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp]\n"
    ret &= "shr eax, 4\n"
    ret &= "and eax, 1\n"
    ret &= "add eax, 48\n"
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "add edi, 4\n"
    ret &= "mov al, 90\n"                             -- Z
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 70\n"                             -- F
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp]\n"
    ret &= "shr eax, 6\n"
    ret &= "and eax, 1\n"
    ret &= "add eax, 48\n"
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "add edi, 4\n"
    ret &= "mov al, 83\n"                             -- S
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 70\n"                             -- F
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp]\n"
    ret &= "shr eax, 7\n"
    ret &= "and eax, 1\n"
    ret &= "add eax, 48\n"
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "add edi, 4\n"
    ret &= "mov al, 84\n"                             -- T
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 70\n"                             -- F
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp]\n"
    ret &= "shr eax, 8\n"
    ret &= "and eax, 1\n"
    ret &= "add eax, 48\n"
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "add edi, 4\n"
    ret &= "mov al, 73\n"                             -- I
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 70\n"                             -- F
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp]\n"
    ret &= "shr eax, 9\n"
    ret &= "and eax, 1\n"
    ret &= "add eax, 48\n"
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "add edi, 4\n"
    ret &= "mov al, 68\n"                             -- D
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 70\n"                             -- F
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp]\n"
    ret &= "shr eax, 10\n"
    ret &= "and eax, 1\n"
    ret &= "add eax, 48\n"
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "add edi, 4\n"
    ret &= "mov al, 79\n"                             -- O
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 70\n"                             -- F
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp]\n"
    ret &= "shr eax, 11\n"
    ret &= "and eax, 1\n"
    ret &= "add eax, 48\n"
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "add edi, 4\n"
    ret &= "mov al, 73\n"                             -- I
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 79\n"                             -- O
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 80\n"                             -- P
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 76\n"                             -- L
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp]\n"
    ret &= "shr eax, 12\n"
    ret &= "and eax, 3\n"
    ret &= "add eax, 48\n"
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "add edi, 4\n"
    ret &= "mov al, 78\n"                             -- N
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 84\n"                             -- T
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp]\n"
    ret &= "shr eax, 14\n"
    ret &= "and eax, 1\n"
    ret &= "add eax, 48\n"
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "add edi, 4\n"
    ret &= "mov al, 82\n"                             -- R
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 70\n"                             -- F
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp]\n"
    ret &= "shr eax, 16\n"
    ret &= "and eax, 1\n"
    ret &= "add eax, 48\n"
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "add edi, 4\n"
    ret &= "mov al, 86\n"                             -- V
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 77\n"                             -- M
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov al, 61\n"                             -- =
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "mov eax, [ebp]\n"
    ret &= "shr eax, 17\n"
    ret &= "and eax, 1\n"
    ret &= "add eax, 48\n"
    ret &= "mov [edi], al\n"
    ret &= "add edi, 2\n"

    ret &= "pop edi\n"
    ret &= "add edi, 160\n"
    ret &= "push edi\n"

    ret &= "mov ebx, #" & sprintf("%x", pressf1) & "\n"
    ret &= "mov dl, 12\n"
    ret &= "call near @@dbg_printtext\n"

    ret &= "pop edi\n"

    ret &= "call near @@dbg_waitkey\n"
    
    ret &= "cmp al, 59\n"
    ret &= "jne @@dbg_nohelp\n"
        ret &= "call near @@dbg_clrscreen\n"
        ret &= "mov ecx, " & sprint(helplines) & "\n"
        ret &= "mov ebx, #" & sprintf("%x", help) & "\n"
        ret &= "mov edi, #B8000\n"
        ret &= "mov dl, 7\n"
        ret &= "@@dbg_helploop:\n"
            ret &= "push edi\n"
            ret &= "call near @@dbg_printtext\n"
            ret &= "pop edi\n"
            ret &= "add edi, 160\n"
            ret &= "dec ecx\n"
        ret &= "jnz @@dbg_helploop:\n"
        ret &= "call near @@dbg_waitkey\n"
        ret &= "jmp near @@dbg_redraw\n"
    ret &= "@@dbg_nohelp:\n"

    ret &= "cmp al, 1\n"
    ret &= "jne @@dbg_noquit\n"
         ret &= "xor eax, eax\n"
         ret &= "mov [#" & sprintf("%x", states[length(states)]) & "], eax\n"
         ret &= "call near @@dbg_clrscreen\n"
    ret &= "@@dbg_noquit:\n"

    ret &= "@@dbg_debugskip:\n"
    ret &= "popf\n"
    ret &= "popad\n"
    ret &= "ret 4\n"

    -- prints a 32-bit number as a hexadecimal number
    -- in: eax(number), edi(pointer to screen pos)
    -- out: none
    -- destroys: ecx, edi, dl
    ret &= "@@dbg_printhex:\n"
    ret &= "mov ecx, 8\n"
    ret &= "add edi, 14\n"
    ret &= "@@dbg_printloop:\n"

        ret &= "mov dl, al\n"
        ret &= "and dl, 15\n"
        ret &= "cmp dl, 10\n"

        ret &= "jae @@dbg_alpha\n"   -- if < 10 then
            ret &= "add dl, 48\n"
            ret &= "mov [edi], dl\n"
            ret &= "sub edi, 2\n"
        ret &= "jmp @@dbg_notalpha\n"
        ret &= "@@dbg_alpha:\n"
            ret &= "add dl, 55\n"
            ret &= "mov [edi], dl\n"
            ret &= "sub edi, 2\n"
        ret &= "@@dbg_notalpha:\n"
        
        ret &= "shr eax, 4\n"
        ret &= "dec ecx\n"
    ret &= "jnz @@dbg_printloop\n"
    ret &= "add edi, 18\n"
    ret &= "ret 0\n"

    -- waits for a keystroke
    -- in: none
    -- out: al (key)
    -- destroys: ax
    ret &= "@@dbg_waitkey:\n"
    ret &= "@@dbg_firstkbloop:\n"
        ret &= "mov ax, [#41A]\n"
        ret &= "mov [#41C], ax\n"
        ret &= "in al, #60\n"
        ret &= "cmp al, 128\n"
    ret &= "jae @@dbg_firstkbloop\n"

    ret &= "@@dbg_secondkbloop:\n"
        ret &= "mov ax, [#41A]\n"
        ret &= "mov [#41C], ax\n"
        ret &= "in al, #60\n"
        ret &= "cmp al, 128\n"
    ret &= "jb @@dbg_secondkbloop\n"
    ret &= "sub al, 128\n"
    ret &= "ret 0\n"

    -- clears the text screen
    -- in: none
    -- out: none
    -- destroys: none
    ret &= "@@dbg_clrscreen:\n"
    ret &= "push edi\n"
    ret &= "push eax\n"
    ret &= "push ecx\n"

    ret &= "mov edi, #B8000\n"
    ret &= "cld\n"
    ret &= "mov ecx, 1000\n"
    ret &= "mov eax, #07000700\n"
    ret &= "rep stosd\n"

    ret &= "pop ecx\n"
    ret &= "pop eax\n"
    ret &= "pop edi\n"

    ret &= "ret 0\n"
    
    -- prints chars, until a \n appears
    -- in: edi (pointer to screen pos), ebx (pointer to text), dl (color)
    -- out: none
    -- destroys: eax, edi, ebx
    ret &= "@@dbg_printtext:\n"
    ret &= "@@dbg_printtextloop:\n"
        ret &= "mov eax, eax\n"                           -- asm.e has got parsing difficulties with the db command after a label, so this is necessary
        ret &= "db #80\ndb #3B\ndb #0A\n"              -- this is the opcode of the next line (commented out)
--        ret &= "cmp [ebx], byte 10\n"          -- asm.e doesn't assemble this right.
        ret &= "jz @@dbg_doneprinttext\n"                   -- done?
        ret &= "mov al, [ebx]\n"
        ret &= "mov [edi], al\n"                -- put char
        ret &= "inc edi\n"
        ret &= "mov [edi], dl\n"
        ret &= "inc edi\n"
        ret &= "inc ebx\n"
        ret &= "jmp @@dbg_printtextloop\n"
    ret &= "@@dbg_doneprinttext:\n"
    ret &= "inc ebx\n"
    ret &= "ret 0\n"

    ret &= "@@dbg_setstate:\n"
    ret &= "mov [#" & sprintf("%x", states[length(states)]) & "], ebx\n"
    ret &= "ret 0\n"

    -- Maybe you wonder why I'm always writing 'ret 0' instead of 'ret'?
    -- Well, my asm.e (modified) has got problems interpreting a 'ret' followed by
    -- a label. The official version of asm.e cannot parse a 'ret 0'.
    -- I need that feature for the main @@dbg_debugger routine.

    return ret
end function

global procedure asm_cleanup()
    for a = 1 to length(pstrucs) do
        free(pstrucs[a])
    end for
    for a = 1 to length(infostrucs) do
        free(infostrucs[a])
    end for
end procedure
